import suite2p.register
import suite2p.nonrigid
import suite2p.run_s2p
import suite2p.dcnv
import suite2p.sparsedetect
import suite2p.sourcery
import suite2p.roiextract
import suite2p.classifier
import suite2p.chan2detect
import suite2p.utils
name = "suite2p"
